@@ -439,6 +439,7 @@ index((jpl_class_to_type), 2, jpl, jpl).
 index((jpl_classname_to_class), 2, jpl, jpl).
 index((jpl_classname_to_type), 2, jpl, jpl).
 index((jpl_datum_to_type), 2, jpl, jpl).
+index((jpl_entityname_to_type), 2, jpl, jpl).
 index((jpl_false), 1, jpl, jpl).
 index((jpl_is_class), 1, jpl, jpl).
 index((jpl_is_false), 1, jpl, jpl).
@@ -457,6 +458,7 @@ index((jpl_ref_to_type), 2, jpl, jpl).
 index((jpl_true), 1, jpl, jpl).
 index((jpl_type_to_class), 2, jpl, jpl).
 index((jpl_type_to_classname), 2, jpl, jpl).
+index((jpl_type_to_entityname), 2, jpl, jpl).
 index((jpl_void), 1, jpl, jpl).
 index((jpl_array_to_length), 2, jpl, jpl).
 index((jpl_array_to_list), 2, jpl, jpl).
@@ -489,6 +491,7 @@ index((portray_clause), 1, prolog_listing, listing).
 index((portray_clause), 2, prolog_listing, listing).
 index((portray_clause), 3, prolog_listing, listing).
 index((member), 2, lists, lists).
+index((memberchk), 2, lists, lists).
 index((append), 2, lists, lists).
 index((append), 3, lists, lists).
 index((prefix), 2, lists, lists).
@@ -705,6 +708,7 @@ index((run_tests), 0, plunit, plunit).
 index((run_tests), 1, plunit, plunit).
 index((load_test_files), 1, plunit, plunit).
 index((running_tests), 0, plunit, plunit).
+index((current_test), 5, plunit, plunit).
 index((test_report), 1, plunit, plunit).
 index((porter_stem), 2, porter_stem, porter_stem).
 index((unaccent_atom), 2, porter_stem, porter_stem).
@@ -1216,6 +1220,10 @@ index((concurrent), 3, thread, thread).
 index((concurrent_maplist), 2, thread, thread).
 index((concurrent_maplist), 3, thread, thread).
 index((concurrent_maplist), 4, thread, thread).
+index((concurrent_forall), 2, thread, thread).
+index((concurrent_forall), 3, thread, thread).
+index((concurrent_and), 2, thread, thread).
+index((concurrent_and), 3, thread, thread).
 index((first_solution), 3, thread, thread).
 index((call_in_thread), 2, thread, thread).
 index((thread_pool_create), 3, thread_pool, thread_pool).
